from __future__ import annotations
from random import Random, randint
from typing import Any, Iterator, List, Callable, Optional
from torch.utils.data import IterableDataset

__version__ = "0.1.0"


class ExtendedIterableDataset(IterableDataset):
    """
    This class extends IterableDataset by allowing data shuffling, applying transformations, and limiting the amount of data.

    Parameters
    ----------
    shuffle_buffer : int
        The size of the shuffle buffer. If it's greater than 1, data is buffered and shuffled before being returned.
    shuffle_seed : Optional[int]
        The seed for the random number generator used for shuffling.
    offset : int
        The number of first data points to skip.
    limit : int
        The maximum number of data points to return. If it's negative, all data is returned.
    transforms : List[Callable[[Any], Any]]
        A list of transformations to apply to the data.
    transforms_required : bool
        If it's true and transforms is empty, an exception is raised.
    *args
        Additional arguments for the IterableDataset.
    **kwargs
        Additional keyword arguments for the IterableDataset.
    """

    def __init__(
        self,
        shuffle_buffer: int = 1,
        shuffle_seed: Optional[int] = None,
        offset: int = 0,
        limit: int = -1,
        transforms: Optional[List[Callable[[Any], Any]]] = None,
        transforms_required: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        """
        Some datasets may use auxiliary classes as singletons to generate data.
        For example, a reusable chess.Board instance.
        
        In order to avoid bugs with its state,
        use the "transforms_required" flag to warn the user
        that data should be extracted from this instance.
        """
        if transforms_required and not transforms:
            raise ValueError(
                f"ExtendedIterableDataset requires transforms in order to avoid bugs. Please read the comment above"
            )

        self.shuffle_buffer = shuffle_buffer
        self.shuffle_seed = (
            shuffle_seed if shuffle_seed is not None else Random().randint(0, 999331)
        )
        self.buffer = []

        self.offset = offset
        self.limit = limit
        self.counter = 0

        self.transforms = transforms if transforms is not None else []

    def generator(self) -> Iterator[Any]:
        """
        Should yield items. This method should be implemented in subclasses.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in a subclass.
        """
        raise NotImplementedError("please implement generator method")

    @property
    def limit_allows_one_more(self) -> bool:
        """
        Checks if the dataset limit allows to yield one more data item.

        Returns
        -------
        bool
            True if the dataset limit allows to yield one more data item, False otherwise.
        """

        if self.limit < 0:
            return True

        return self.counter < self.limit + self.offset

    def transform_x(self, x: Any) -> Any:
        """
        Applies transformations to the data.

        Parameters
        ----------
        x : Any
            The data point to be transformed.

        Returns
        -------
        Any
            The transformed data point.
        """
        for transform in self.transforms:
            x = transform(x)

        return x

    def flush_buffer(self) -> Iterator[Any]:
        """
        Shuffles and yields the buffer content, then resets the buffer.

        Returns
        -------
        Iterator[Any]
            An iterator over the shuffled buffer content.
        """

        Random(self.shuffle_seed).shuffle(self.buffer)

        for x in self.buffer:
            yield x

        self.buffer = []

    def generator_with_conditions(self) -> Iterator[Any]:
        """
        Generates items taking into account the dataset limit and offset.
        Applies transformations to each generated item.

        Returns
        -------
        Iterator[Any]
            An iterator over the generated and transformed items.

        Raises
        ------
        StopIteration
            If the dataset limit does not allow to yield more items.
        """

        gen = self.generator()

        while True:
            try:
                if not self.limit_allows_one_more:
                    raise StopIteration

                x = next(gen)

                if self.counter < self.offset:
                    self.counter += 1
                    continue

                yield self.transform_x(x)

                self.counter += 1

            except StopIteration:
                break

    def generator_with_buffer(self) -> Iterator[Any]:
        """
        Generates items with a shuffle buffer.
        If the buffer size is greater than 1, items are buffered and shuffled before being yielded.
        If the buffer size is 1 or less, items are yielded as they are generated.

        Returns
        -------
        Iterator[Any]
            An iterator over the generated, possibly shuffled items.
        """

        if self.shuffle_buffer > 1:
            for x in self.generator_with_conditions():
                self.buffer.append(x)

                if len(self.buffer) >= self.shuffle_buffer:
                    yield from self.flush_buffer()

            if len(self.buffer) > 0:
                yield from self.flush_buffer()

        else:
            for x in self.generator_with_conditions():
                yield x

    def __iter__(self) -> Iterator[Any]:
        """
        Generates items with a shuffle buffer.
        If the buffer size is greater than 1, items are buffered and shuffled before being yielded.
        If the buffer size is 1 or less, items are yielded as they are generated.

        Returns
        -------
        Iterator[Any]
            An iterator over the generated, possibly shuffled items.
        """
        return self.generator_with_buffer()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['torch_exid']

package_data = \
{'': ['*']}

install_requires = \
['torch>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'torch-exid',
    'version': '0.1.1',
    'description': 'An extension of PyTorch IterableDataset, this package introduces functionalities for shuffling, limiting, and offsetting data.',
    'long_description': '# Extended Iterable Dataset for PyTorch\nAn extension of PyTorch [IterableDataset](https://pytorch.org/docs/stable/data.html#torch.utils.data.IterableDataset), this package introduces functionalities for shuffling, limiting, and offsetting data.\n\n## Installation\n\nDirectly from [PyPI](https://pypi.org/project/torch_exid/):\n```bash\npip install torch_exid\n```\n\nOr using [Poetry](https://python-poetry.org/):\n```bash\npoetry add torch_exid\n```\n\n## Usage\nBegin by subclassing `ExtendedIterableDataset` and implement the `generator` method to yield items.\n\nHere\'s a simple example using an `IntegersDataset`:\n```python\nfrom torch_exid import ExtendedIterableDataset\n\nclass IntegersDataset(ExtendedIterableDataset):\n    def generator(self) -> Iterator[int]:\n        n = 0\n        while True:\n            yield n\n            n += 1\n\n# Will print out integers 0, 1, ..., 9:\nfor n in IntegersDataset(limit=10):\n    print(n)\n```\n\n## Constructor Parameters\n`ExtendedIterableDataset` introduces several parameters to provide additional control:\n\n### limit: int\nSets the maximum number of data points to return. If negative, all data points are returned. Default is `-1` (return all data).\n```python\n# Will print out "0, 1, 2"\nfor n in IntegersDataset(limit=3)\n    print(n)\n```\n\n### offset: int\nDetermines the number of initial data points to skip. Default is `0`.\n```python\n# Will print out "2, 3, 4"\nfor n in IntegersDataset(limit=3, offset=2)\n    print(n)\n```\n\n### shuffle_buffer: int\nThis specifies the buffer size for shuffling. If greater than `1`, data is buffered and shuffled prior to being returned. If set to `1` (default), no shuffling occurs.\n\n```python\n# Will print out "0, 1, 3, 2" for the first time...\nfor n in IntegersDataset(limit=4, shuffle_buffer=2)\n    print(n)\n\n# ...and 1, 0, 2, 3 second time\nfor n in IntegersDataset(limit=4, shuffle_buffer=2)\n    print(n)\n```\n\n### shuffle_seed: int\nDefines the seed for the random number generator used in shuffling. If not provided, a random seed is used:\n\n```python\n# Will print out "1, 0, 3, 2" both times:\nfor n in IntegersDataset(limit=4, shuffle_buffer=2, shuffle_seed=42)\n    print(n)\n\nfor n in IntegersDataset(limit=4, shuffle_buffer=2, shuffle_seed=42)\n    print(n)\n```\n\n### transforms: List[Callable[[Any], Any]]\nA list of transformations to apply to the data. Default is an empty list.\n```python\nds = IntegersDataset(\n    limit=3,\n    transforms=[\n        lambda n: n + 1,\n        lambda n: n ** 2,\n    ],\n)\n\n# Will print out "1, 4, 9"\nfor n in ds:\n    print(n)\n```\n\nIn addition to the above, any arguments or keyword arguments for the [IterableDataset](https://pytorch.org/docs/stable/data.html#torch.utils.data.IterableDataset) superclass can also be passed.\n\n## Contributing\nContributions are greatly appreciated! Improvement can be made by submitting issues, proposing new features, or submitting pull requests with bug fixes or new functionalities.\n\n### Getting started with contributing\nHere are the steps to get started with development:\n```bash\n# Clone the repository:\ngit clone https://github.com/arlegotin/torch_exid.git\ncd torch_exid\n\n# Install the project and its dependencies using Poetry:\npoetry install\n\n# Spawn a shell within the virtual environment:\npoetry shell\n\n# Run tests to ensure everything is working correctly:\npytest tests/\n```\n\nPlease ensure all changes are accompanied by relevant unit tests, and that all tests pass before submitting a pull request. This helps maintain the quality and reliability of the project.\n',
    'author': 'Artem Legotin',
    'author_email': 'hello@artemlegotin.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
